from functools import wraps
from typing import Any, Callable

from .asyncev import AsyncEv, BaseEvent as Event, Listener

_default = AsyncEv()


@wraps(_default.bind)
def bind(event: type[Event], func: Listener):
    _default.bind(event, func)


@wraps(_default.unbind)
def unbind(event: type[Event], func: Listener):
    _default.unbind(event, func)


@wraps(_default.emit)
def emit(event: Event):
    _default.emit(event)


@wraps(_default.gather)
async def gather(event: Event) -> list[Any]:
    return await _default.gather(event)


@wraps(_default.gather_for)
def gather_for(event: Event, func: Callable[..., None]):
    _default.gather_for(event, func)


@wraps(_default.wait_for)
async def wait_for(event: type[Event]) -> Event:
    return await _default.wait_for(event)


__all__ = [
    "AsyncEv",
    "Event",
    "Listener",
    "bind",
    "unbind",
    "emit",
    "gather",
    "gather_for",
    "wait_for",
]



from abc import ABC
import asyncio
import logging
from dataclasses import dataclass
from inspect import iscoroutinefunction, ismethod
from collections.abc import Coroutine
from typing import Any, Callable, Optional, TypeVar, Union
from weakref import ReferenceType, WeakMethod, ref


log = logging.getLogger(__name__)


class AsyncEvError(Exception):
    pass

@dataclass
class BaseEvent(ABC):
    def __post_init__(self):
        if type(self) is BaseEvent:
            raise Exception("BaseEvent should not be instantiated directly")

class SENTINEL(BaseEvent):
    pass

EventType = TypeVar("EventType", bound=BaseEvent)
Listener = Callable[[BaseEvent], Coroutine[None, None, Any]]
Reference = Union[ReferenceType[Listener], WeakMethod[Listener]]


def funcref(
    listener: Listener,
    callback: Optional[Callable[[Reference], None]] = None,
) -> Reference:
    """Unified way to create a reference to a function or a bound method."""
    if ismethod(listener):
        return WeakMethod(listener, callback)
    return ref(listener, callback)


class AsyncEv:
    """
    AsyncEv uses the asyncio event loop to create Tk-style named events.
    Events can be emitted, awaited, and bound to.
    """

    def __init__(self):
        self.events: dict[type[BaseEvent], set[Reference]] = {}
        self.writelock: dict[type[BaseEvent], asyncio.Lock] = {}
        # FIXME: Writelock can't be used in non-async methods...

    def bind(self, event: type[BaseEvent], listener: Listener):
        """Bind {listener} to a named {event}."""
        if not iscoroutinefunction(listener):
            raise TypeError("Event listeners must be async!")

        log.debug("Adding binding: %r -> %s", event, listener)
        if event not in self.events:
            self.events[event] = set()
            self.writelock[event] = asyncio.Lock()

        self.events[event].add(
            funcref(listener, lambda r: self.events[event].discard(r))
        )

    def unbind(self, event: type[BaseEvent], listener: Listener):
        """Unbind {listener} from {event}.
        Will error if event or listener doesn't exist."""
        log.debug("Removing binding: %r -> %s", event, listener)

        self.events[event].discard(funcref(listener))

    async def _emit(self, event: BaseEvent):
        """Call all valid {event} handlers with the provided events."""
        t = type(event)
        async with self.writelock[t]:
            # the weakref callback _should_ ensure that no Nones are left in the list.
            callbacks = [task()(event) for task in self.events[t]] # type: ignore

        log.debug("Emit %s!", t.__name__)
        return await asyncio.gather(*callbacks)

    def emit(self, event: BaseEvent):
        """Emit {event}, triggering all listeners as soon as possible."""
        log.debug("Scheduling emit(%s)", event)
        asyncio.create_task(self._emit(event))

    async def gather(self, event: BaseEvent) -> list[Any]:
        """emit {event}, wait for all listeners to run, and return the result."""

        log.debug("Gather(%s)", event)
        return await self._emit(event)

    def gather_for(self, event: BaseEvent, func: Callable[..., None]):
        """Emit {event}, wait for all listeners, and send their results to {func}.
        This is useful for passing results to synchronous functions.
        """

        async def _callback():
            results = await self.gather(event)
            func(results)

        log.debug("Scheduling gather_for(%s)", event)
        asyncio.create_task(_callback())

    async def wait_for(self, event: type[BaseEvent]) -> BaseEvent:
        """Sleep until {event} occurs.

        wait_for creates a temporary handler for the given event that gets
        removed after the event has occurred once. This is more expensive
        than registering a proper event handler, so should probably only be
        used for events that don't occur very often.
        """
        ev = asyncio.Event()

        out: Union[BaseEvent, SENTINEL] = SENTINEL()

        async def wait_trigger(event: BaseEvent):
            nonlocal out
            out = event
            ev.set()

        self.bind(event, wait_trigger)
        await ev.wait()
        self.unbind(event, wait_trigger)
        if type(out) is SENTINEL:
            raise AsyncEvError("wait_for triggered without passing an event! this is a bug!")
        return out



from dataclasses import dataclass
from typing import Any
from weakref import ref, WeakMethod
import asyncio
import logging
import asyncev

from unittest import IsolatedAsyncioTestCase

yield_for = 0.01  # how long to allow the event handler to run after setup

log = logging.getLogger(__name__)


@dataclass
class ValueEvent(asyncev.Event):
    value: Any


class EventTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.eventhandler = asyncev.AsyncEv()
        self.arg_value = None
        self.method_called = False
        self.arg_input = "argument value"

    async def test_bound_function(self):
        """Test that
        * a function bound to an event receives events of that type,
        * unbinding removes the function from the listeners and no longer receives events.
        """

        # check that there are no listeners for our event
        self.assertNotIn(ValueEvent, self.eventhandler.events)

        async def function(ev: ValueEvent):
            self.arg_value = ev.value

        self.eventhandler.bind(ValueEvent, function)

        # check that there is now a bound listener for the event.
        self.assertIn(ValueEvent, self.eventhandler.events)
        self.assertIn(ref(function), self.eventhandler.events[ValueEvent])

        # check that the before-value is correct
        self.assertIsNone(self.arg_value)

        # check that the bound listener receives the event.
        self.eventhandler.emit(ValueEvent(self.arg_input))
        await asyncio.sleep(yield_for)
        self.assertEqual(self.arg_input, self.arg_value)

        # check that unbinding makes the event not arrive to the listener.
        self.eventhandler.unbind(ValueEvent, function)
        self.arg_value = None
        self.eventhandler.emit(ValueEvent(self.arg_input))
        await asyncio.sleep(yield_for)
        self.assertIsNone(self.arg_value)

    async def test_bound_method(self):
        """Same as test_bound_function, but for methods."""

        # check that there are no listeners for our event
        self.assertNotIn(ValueEvent, self.eventhandler.events)

        # set up an class that binds to an event on instantiation
        class BindObject:
            def __init__(this, eventhandler):
                eventhandler.bind(ValueEvent, this.listener)
                this.arg_value = None

            async def listener(self, ev: ValueEvent):
                self.arg_value = ev.value

        obj = BindObject(self.eventhandler)

        # check that there is now a bound listener for the event.
        self.assertIn(ValueEvent, self.eventhandler.events)
        self.assertIn(WeakMethod(obj.listener), self.eventhandler.events[ValueEvent])

        # check that the before-value is correct
        self.assertIsNone(self.arg_value)

        # check that the bound listener receives the event.
        self.eventhandler.emit(ValueEvent(self.arg_input))
        await asyncio.sleep(yield_for)
        self.assertEqual(self.arg_input, obj.arg_value)

        # check that unbinding makes the event not arrive to the listener.
        self.eventhandler.unbind(ValueEvent, obj.listener)
        self.arg_value = None
        self.eventhandler.emit(ValueEvent(self.arg_input))
        await asyncio.sleep(yield_for)
        self.assertIsNone(self.arg_value)

    async def test_wait_for(self):
        """test that wait_for cleans up after itself"""

        async def waiter():
            self.method_called = True
            ev = await self.eventhandler.wait_for(ValueEvent)
            self.arg_value = ev.value

        # check that there are no registered listeners for the event
        self.assertNotIn(ValueEvent, self.eventhandler.events)
        # check before-values
        self.assertFalse(self.method_called)
        self.assertNotEqual(self.arg_input, self.arg_value)

        # check that the listener has started and is blocking
        task = asyncio.create_task(waiter())
        await asyncio.sleep(yield_for)
        self.assertTrue(self.method_called)
        self.assertNotEqual(self.arg_input, self.arg_value)

        # check that the listener has received the event
        self.eventhandler.emit(ValueEvent(self.arg_input))
        await asyncio.sleep(yield_for)
        self.assertEqual(self.arg_value, self.arg_input)

        # check that the wait_for ephemeral listener has been removed
        self.assertFalse(self.eventhandler.events[ValueEvent])
        await task  # clean up waiter

    async def test_gather(self):
        """test that gather returns all results"""

        def factory(i: int):
            async def f(ev: ValueEvent) -> int:
                return i + ev.value

            return f

        funcs = []
        for i in range(10):
            f = factory(i)
            funcs.append(f)
            self.eventhandler.bind(ValueEvent, f)

        results = await self.eventhandler.gather(ValueEvent(10))
        self.assertEqual(sorted(results), list(range(10, 20)))

    async def test_gather_for(self):
        def factory(i: int):
            async def f(ev: ValueEvent):
                return i + ev.value

            return f

        funcs = []
        for i in range(10):
            f = factory(i)
            funcs.append(f)
            self.eventhandler.bind(ValueEvent, f)

        def gatherer(results):
            self.method_called = True
            self.assertEqual(sorted(results), list(range(10, 20)))

        self.eventhandler.gather_for(ValueEvent(10), gatherer)
        await asyncio.sleep(yield_for)

        self.assertTrue(self.method_called)

    async def test_prune(self):
        """Check that listeners are cleaned up when they go out of scope"""
        class BindObject:
            def __init__(this, eventhandler):
                eventhandler.bind(ValueEvent, this.bound_method)
                this.arg_value = None

            async def bound_method(self, ev: ValueEvent):
                self.arg_value = ev.value

        # check that there are no listeners already
        self.assertNotIn(ValueEvent, self.eventhandler.events)

        obj = BindObject(self.eventhandler)

        # check that our listener has been added
        self.assertIn(ValueEvent, self.eventhandler.events)
        self.assertIn(
            WeakMethod(obj.bound_method), self.eventhandler.events[ValueEvent]
        )

        del obj
        # check that the listener was immediately deleted
        self.assertSetEqual(self.eventhandler.events[ValueEvent], set())

    def test_non_async_handler(self):
        def non_async_handler(ev: ValueEvent):
            pass

        with self.assertRaises(TypeError):
            self.eventhandler.bind(ValueEvent, non_async_handler)
